from .parser import parse, ParserError
from .query import mongodb


def convert_to_mongo(rql_str):
    try:
        return mongodb.unparse(parse(rql_str))
    except ValueError:
        raise ParserError('query not valid: {}'.format(rql_str))



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function

"""
This module provides the same interface as mediajson for purposes
of backwards compatibility. It is recommended that new code uses
that library instead.
"""

from mediajson import dump, dumps, load, loads
from mediajson import encode_value, decode_value
from mediajson import JSONEncoder, JSONDecoder
from mediajson import NMOSJSONEncoder, NMOSJSONDecoder

__all__ = ["dump", "dumps", "load", "loads",
           "encode_value", "decode_value",
           "JSONEncoder", "JSONDecoder",
           "NMOSJSONEncoder", "NMOSJSONDecoder"]



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netifaces
from .mdnsInterface import MDNSInterface
from nmoscommon import nmoscommonconfig
from nmoscommon.mdns.mdnsExceptions import NoNetworkInterfacesFoundException
try:
    from pyipputils import ipphostname
    ippUtilsAvailable = True
except ImportError:
    ippUtilsAvailable = False
    ipphostname = None  # used to make test patching work


class MDNSInterfaceController(object):

    def __init__(self, logger):
        self.logger = logger
        self.interfaces = {}
        self.defaultInterfaces = []
        self._get_default_interfaces()

    def addInterface(self, ip):
        if ip not in self.interfaces.keys():
            self.interfaces[ip] = MDNSInterface(ip)
        return self.interfaces[ip]

    def getInterfaces(self, interfaceIps):
        if interfaceIps == []:
            return self.defaultInterfaces
        else:
            interfaceList = []
            for ip in interfaceIps:
                interfaceList.append(self.addInterface(ip))
                self.logger.writeDebug("Switching to using mDNS interface provided:{}".format(ip))
            return interfaceList

    def close(self):
        for key, interface in self.interfaces.items():
            interface.close()

    def _get_default_interfaces(self):
        interfaces = self._get_default_interface_from_file()
        if not interfaces:
            interfaces = self._get_default_interface_from_routing_rules()
        if not interfaces:
            interfaces = self._get_default_interface_from_gateway()
        if not interfaces:
            interfaces = self._get_all_interfaces()
        if not interfaces:
            msg = "Could not find any interfaces for mDNS use"
            self.logger.writeError(msg)
            raise NoNetworkInterfacesFoundException(msg)
        for interface in interfaces:
            self.defaultInterfaces.append(self.addInterface(interface))

    def _get_all_interfaces(self):
        toReturn = []
        for interface in netifaces.interfaces():
            interfaceDetails = netifaces.ifaddresses(interface)
            try:
                ipv4Addr = interfaceDetails[netifaces.AF_INET][0]['addr']
            except KeyError:
                pass
            toReturn.append(ipv4Addr)
        if toReturn == []:
            return False
        else:
            self.logger.writeDebug("Using all available IPv4 network interfaces for mDNS:"
                                   "{}".format(toReturn))
        return toReturn

    def _get_default_interface_from_routing_rules(self):
        if ippUtilsAvailable:
            interfaces = [ipphostname.getLocalIP()]
            self.logger.writeDebug("Choosing mDNS interface using routing rules:"
                                   "{}".format(interfaces[0]))
            return interfaces
        else:
            self.logger.writeDebug("Could not find ipp-utils, will try using default gateway interface")
            return False

    def _get_default_interface_from_gateway(self):
        defaultGateways = netifaces.gateways()
        try:
            defaultInterfaceName = defaultGateways[netifaces.AF_INET][0][1]
        except KeyError:
            return False
        defaultInterfaceDetails = netifaces.ifaddresses(defaultInterfaceName)
        defaultInterfaceIp = defaultInterfaceDetails[netifaces.AF_INET][0]['addr']
        self.logger.writeDebug("Choosing mDNS interface using the default gateway:"
                               " {}".format(defaultInterfaceIp))
        return [defaultInterfaceIp]

    def _get_default_interface_from_file(self):
        try:
            interfaces = []
            interfaceNames = nmoscommonconfig.config['interfaces']
            if interfaceNames == "*":
                return self._get_all_interfaces()
            for interfaceName in interfaceNames:
                interfaces.append(netifaces.ifaddresses(interfaceName)[netifaces.AF_INET][0]['addr'])
            self.logger.writeDebug("Choosing mDNS interface from /etc/nmoscommon/config.json "
                                   "file: {}".format(interfaces))
        except KeyError:
            self.logger.writeDebug("No interface config file - will try and use routing rules")
            return False
        else:
            return interfaces



#!/usr/bin/env python
from zeroconf_monkey import ServiceBrowser


class MDNSSubscription(object):

    def __init__(self, zeroconf, listener, regtype):
        self.listener = listener
        self.regtype = regtype + ".local."
        self.zeroconf = zeroconf
        self.browser = ServiceBrowser(self.zeroconf, self.regtype, self.listener)

    def close(self):
        self.browser.cancel()



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .mdnsExceptions import ServiceAlreadyExistsException, ServiceNotFoundException


class MDNSRegistrationController(object):

    def __init__(self):
        self.registrations = {}

    def addRegistration(self, registration):
        name = registration.name
        regtype = registration.regtype
        if registration.regtype in self.registrations:
            if registration.name in self.registrations[regtype]:
                raise ServiceAlreadyExistsException()
        else:
            self.registrations[regtype] = {}
        self.registrations[regtype][name] = registration
        registration.register()

    def removeRegistration(self, name, regtype):
        try:
            self.registrations[regtype].pop(name).unRegister()
        except KeyError:
            raise ServiceNotFoundException
        if self.registrations[regtype] == {}:
            self.registrations.pop(regtype)

    def close(self):
        for regtype in self.registrations:
            for name in self.registrations[regtype]:
                self.registrations[regtype][name].unRegister()
        self.registrations = {}



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function

from six import string_types
from six import iterkeys
from six import get_method_self

import uuid
import json
import traceback
import time

from flask import Flask, Response, request, abort
from flask_sockets import Sockets
from nmoscommon.flask_cors import crossdomain
from functools import wraps
import re

from pygments import highlight
from pygments.lexers import JsonLexer, PythonTracebackLexer
from pygments.formatters import HtmlFormatter
from mediajson import NMOSJSONEncoder
import sys

from werkzeug.exceptions import HTTPException
from werkzeug.wrappers import BaseResponse

# This moved in Werkzeug 0.15.0, but if it's not there try the old location in case we're on Bionic with 0.14.1
try:
    from werkzeug.middleware.proxy_fix import ProxyFix
    HAS_WERKZEUG_MIDDLEWARE = True
except ImportError:
    from werkzeug.contrib.fixers import ProxyFix
    HAS_WERKZEUG_MIDDLEWARE = False

from requests.structures import CaseInsensitiveDict

import flask_oauthlib
import flask_oauthlib.client
from authlib.common.errors import AuthlibBaseError, AuthlibHTTPError

from .utils import getLocalIP

try:  # pragma: no cover
    from six.moves.urllib.parse import urlparse
    import urllib2 as http
except Exception:  # pragma: no cover
    from urllib import request as http
    from urllib.parse import urlparse

from nmoscommon.nmoscommonconfig import config as _config


class MediaType(object):
    def __init__(self, mr):
        comps = [x.strip() for x in mr.split(';')]
        (self.type, self.subtype) = comps.pop(0).split('/')
        self.priority = 1.0
        self.options = []
        self.accept_index = 0
        for c in comps:
            if c[0] == 'q':
                tmp = [x.strip() for x in c.split('=')]
                if tmp[0] == "q":
                    self.priority = float(tmp[1])
                    continue
            self.options.append(c)

    def __str__(self):
        return "%s/%s;%s" % (self.type, self.subtype, ';'.join(self.options + ['q=%g' % self.priority, ]))

    def __repr__(self):
        return "MediaType(\"{}\")".format(self.__str__())

    def matches(self, t):
        if not isinstance(t, MediaType):
            t = MediaType(t)
        if t.type == self.type or self.type == '*':
            if t.subtype == self.subtype or self.subtype == '*':
                return True
        return False


def AcceptStringParser(accept_string):
    return [MediaType(x.strip()) for x in accept_string.split(',')]


def AcceptableType(mt, accept_string):
    if not isinstance(mt, MediaType):
        mt = MediaType(mt)
    for idx, t in enumerate(AcceptStringParser(accept_string)):
        if t.matches(mt):
            t.accept_index = idx
            return t
    return None


def MostAcceptableType(type_strings, accept_string):
    """First parameter is a list of media type strings, the second is an HTTP Accept header string. Return
    value is a string which corresponds to the most acceptable type out of the list provided."""
    def __cmp(a, b):
        if a is None and b is None:
            return 0
        elif a is None:
            return 1
        elif b is None:
            return -1
        elif a.priority == b.priority:
            return (a.accept_index > b.accept_index) - (a.accept_index < b.accept_index)
        else:
            return -((a.priority > b.priority) - (a.priority < b.priority))

    return sorted(
                [(mt, AcceptableType(MediaType(mt), accept_string)) for mt in type_strings],
                cmp=lambda a, b: __cmp(
                    a[1],
                    b[1]
                )
          )[0][0]


HOST = None
itt = 0

# Sometimes interfaces don't come up in time
# during boot, wait until we find an IP or fail
# after 5 seconds
while not HOST:
    try:
        HOST = getLocalIP()
    except Exception:
        if itt > 5:
            raise OSError("Could not find an interface for webapi")
        else:
            time.sleep(1)
            itt = itt + 1


class LinkingHTMLFormatter(HtmlFormatter):
    def wrap(self, source, outfile):
        return self._wrap_linking(super(LinkingHTMLFormatter, self).wrap(source, outfile))

    def _wrap_linking(self, source):
        for i, t in source:
            m = re.match(r'(.*)&quot;([a-zA-Z0-9_.\-~]+)/&quot;(.+)', t)
            if m:
                t = m.group(1) + '&quot;<a href="./' + m.group(2) + '/">' + m.group(2) + '/</a>&quot;' + m.group(3) + "<br>"

            yield i, t


def expects_json(func):
    @wraps(func)
    def __inner(ws, msg, **kwargs):
        return func(ws, json.loads(msg), **kwargs)
    return __inner


def htmlify(r, mimetype, status=200):
    # if the request was proxied via the nodefacade, use the original host in response.
    # additional external proxies could cause an issue here, so we can override the hostname
    # in the config to an externally-defined one if there are multiple reverse proxies
    path = request.headers.get('X-Forwarded-Path', request.path)
    host = _config.get('node_hostname')
    if host is None:
        host = request.headers.get('X-Forwarded-Host', request.host)
    if _config.get('https_mode') == 'enabled':
        scheme = 'https'
    else:
        scheme = request.headers.get('X-Forwarded-Proto', urlparse(request.url).scheme)
    base_url = "{}://{}".format(scheme, host)

    title = '<a href="' + base_url + '/">' + base_url + '</a>'
    t = base_url
    for x in path.split('/'):
        if x != '':
            t += '/' + x.strip('/')
            title += '/<a href="' + t + '">' + x.strip('/') + '</a>'
    return IppResponse(highlight(json.dumps(r, indent=4, cls=NMOSJSONEncoder),
                                 JsonLexer(),
                                 LinkingHTMLFormatter(linenos='table',
                                                      full=True,
                                                      title=title)),
                       mimetype='text/html',
                       status=status)


def jsonify(r, status=200, headers=None):
    if headers is None:
        headers = {}
    return IppResponse(json.dumps(r, indent=4, cls=NMOSJSONEncoder),
                       mimetype='application/json',
                       status=status,
                       headers=headers)


def returns_response(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        r = f(*args, **kwargs)

        response = r.response
        status = r.status_code
        headers = dict(r.headers)
        mimetype = r.mimetype
        content_type = r.content_type
        direct_passthrough = True

        return IppResponse(response=response,
                           status=status,
                           headers=headers,
                           mimetype=mimetype,
                           content_type=content_type,
                           direct_passthrough=direct_passthrough)
    return decorated_function


def returns_json(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        r = f(*args, **kwargs)
        status = 200
        flatfmt = False
        headers = {}
        if isinstance(r, tuple) and len(r) > 1:
            status = r[0]
            if len(r) > 2:
                headers = r[2]
            r = r[1]
        if isinstance(r, list):
            flatfmt = True
        if isinstance(r, dict) or flatfmt:
            try:
                bm = request.accept_mimetypes.best_match(['application/json', 'text/html', 'rick/roll'])
                if bm in ['text/html', 'rick/roll']:
                    return htmlify(r, bm, status=status)
                else:
                    return jsonify(r, status=status, headers=headers)
            except TypeError:
                if status == 200:
                    status = 204
                return IppResponse(status=status)
        elif isinstance(r, IppResponse):
            return r
        else:
            return IppResponse(r, status=status, headers=headers)
    return decorated_function


def returns_requires_auth(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        r = f(*args, **kwargs)

        if get_method_self(f)._oauth_config is not None:
            if 'token' not in request.headers:
                print('Could not find token')
                return IppResponse(status=401)

            token = request.headers.get('token')
            if not get_method_self(f)._authenticate(token):
                print('Authentication Failed')
                return IppResponse(status=401)

            if not get_method_self(f)._authorize(token):
                print('Authorization Failed.')
                return IppResponse(status=401)

        status = 200
        flatfmt = False
        headers = {'Access-Control-Allow-Credentials': 'true'}
        if isinstance(r, tuple) and len(r) > 1:
            status = r[0]
            if len(r) > 2:
                for h in r[2]:
                    headers[h] = r[2][h]
            r = r[1]
        if isinstance(r, list):
            flatfmt = True
        if isinstance(r, dict) or flatfmt:
            try:
                bm = request.accept_mimetypes.best_match(['application/json', 'text/html', 'rick/roll'])
                if bm in ['text/html', 'rick/roll']:
                    return htmlify(r, bm, status=status)
                else:
                    return jsonify(r, status=status, headers=headers)
            except TypeError:
                if status == 200:
                    status = 204
                return IppResponse(status=status)
        elif isinstance(r, BaseResponse):
            headers = r.headers
            headers['Access-Control-Allow-Credentials'] = 'true'
            if isinstance(r, IppResponse):
                return r
            else:
                return IppResponse(response=r.response,
                                   status=r.status,
                                   headers=headers,
                                   mimetype=r.mimetype,
                                   content_type=r.content_type,
                                   direct_passthrough=r.direct_passthrough)
        else:
            return IppResponse(response=r, status=status, headers=headers)

    return decorated_function


def returns_file(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        r = f(*args, **kwargs)
        if 'content-type' in r and r['content-type'] is not None:
            return IppResponse(r['content'],
                               status=200,
                               headers={'Content-Disposition': 'attachment; filename={}.{}'.format(r['filename'], r['type'])},
                               content_type=r['content-type'])
        else:
            return IppResponse(r['content'],
                               status=200,
                               headers={'Content-Disposition': 'attachment; filename={}.{}'.format(r['filename'], r['type'])})

    return decorated_function


def obj_path_access(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        path = kwargs.pop('path', '')
        rep = f(*args, **kwargs)
        if request.method == "GET":
            p = [x for x in path.split('/') if x != '']
            for x in p:
                if isinstance(rep, dict):
                    if x not in rep:
                        abort(404)
                    rep = rep[x]
                elif isinstance(rep, list) or isinstance(rep, tuple):
                    try:
                        y = int(x)
                    except Exception:
                        abort(404)
                    if y < 0 or y >= len(rep):
                        abort(404)
                    rep = rep[y]
                else:
                    abort(404)
            try:
                repdump = json.dumps(rep)
            except Exception:
                abort(404)
            if repdump == "{}":
                return []
            else:
                if isinstance(rep, string_types):
                    """Returning a string here will bypass the jsonification, which we want, have to do it manually"""
                    return jsonify(rep)
                else:
                    return rep
        else:
            return rep
    return decorated_function


def secure_route(path, methods=None, auto_json=True, headers=None, origin='*'):
    if methods is None:
        methods = ['GET', 'HEAD']
    if headers is None:
        headers = []

    def annotate_function(func):
        func.common_path = path
        func.secure_route = path
        func.app_methods = methods
        func.app_auto_json = auto_json
        func.app_headers = headers
        func.app_origin = origin
        return func
    return annotate_function


def basic_route(path):
    def annotate_function(func):
        func.common_path = path
        func.response_route = path
        return func
    return annotate_function


def route(path, methods=None, auto_json=True, headers=None, origin='*'):
    if headers is None:
        headers = []

    def annotate_function(func):
        func.common_path = path
        func.app_route = path
        func.app_methods = methods
        func.app_auto_json = auto_json
        func.app_headers = headers
        func.app_origin = origin
        return func
    return annotate_function


def resource_route(path, methods=None):
    def annotate_function(func):
        func.common_path = path
        func.app_resource_route = path
        func.app_methods = methods
        return func
    return annotate_function


def file_route(path, methods=None, headers=None):
    def annotate_function(func):
        func.common_path = path
        func.app_file_route = path
        func.app_methods = methods
        func.app_headers = headers
        return func
    return annotate_function


def errorhandler(*args, **kwargs):
    def annotate_function(func):
        func.errorhandler_args = args
        func.errorhandler_kwargs = kwargs
        return func
    return annotate_function


def on_json(path):
    def annotate_function(func):
        func.socket_path = path
        return func
    return annotate_function


def wrap_val_in_grain(pval):
    egrain = {
            "grain": {
                "duration": {
                    "denominator": 1,
                    "numerator": 0
                    },
                "flow_id": "00000000-0000-0000-0000-000000000000",
                "grain_type": "event",
                "source_version": 0,
                "rate": {
                    "denominator": 1,
                    "numerator": 0
                    },
                "origin_timestamp": "0:0",
                "source_id": "00000000-0000-0000-0000-000000000000",
                "sync_timestamp": "0:0",
                "creation_timestamp": "0:0",
                "event_payload": {
                    "type": "urn:x-ipstudio:format:event.param.change",
                    "topic": "",
                    "data": [
                        {
                            "path": {},
                            "pre": {},
                            "post": {}
                        }
                    ]
                },
                "flow_instance_id": "00000000-0000-0000-0000-000000000000"
                },
            "@_ns": "urn:x-ipstudio:ns:0.1"
            }
    egrain["grain"]["event_payload"]["data"][0]["post"] = pval
    return json.dumps(egrain)


def grain_event_wrapper(func):
    @wraps(func)
    def wrapper(ws, message):
        pval = func(ws, message["grain"]["event_payload"]["data"][0]["post"])
        if pval is not None:
            ws.send(wrap_val_in_grain(pval))
    return wrapper


class IppResponse(Response):
    def __init__(self, response=None, status=None, headers=None, mimetype=None, content_type=None, direct_passthrough=False):
        headers = CaseInsensitiveDict(headers) if headers is not None else None
        if response is not None and isinstance(response, BaseResponse) and response.headers is not None:
            headers = CaseInsensitiveDict(response.headers)

        if headers is None:
            headers = CaseInsensitiveDict()

        h = headers
        h['Access-Control-Allow-Origin'] = headers.get('Access-Control-Allow-Origin', '*')
        h['Access-Control-Allow-Methods'] = headers.get('Access-Control-Allow-Methods', "GET, PUT, POST, HEAD, OPTIONS, DELETE")
        h['Access-Control-Max-Age'] = headers.get('Access-Control-Max-Age', "21600")
        h['Cache-Control'] = headers.get('Cache-Control', "no-cache, must-revalidate, no-store")
        if 'Access-Control-Allow-Headers' not in headers and len(headers.keys()) > 0:
            h['Access-Control-Allow-Headers'] = ', '.join(iterkeys(headers))

        data = None
        if response is not None and isinstance(response, string_types):
            data = response
            response = None

        if response is not None and isinstance(response, BaseResponse):
            new_response_headers = CaseInsensitiveDict(response.headers if response.headers is not None else {})
            new_response_headers.update(h)
            response.headers = new_response_headers
            headers = None
            data = response.get_data()

        else:
            headers.update(h)
            headers = dict(headers)

        super(IppResponse, self).__init__(response=response,
                                          status=status,
                                          headers=headers,
                                          mimetype=mimetype,
                                          content_type=content_type,
                                          direct_passthrough=direct_passthrough)
        if data is not None:
            self.set_data(data)

    def __eq__(self, other):

        headers_match = True
        for (hdr, val) in self.headers:
            if hdr not in other.headers.keys():
                print("{} not in other".format(hdr))
                headers_match = False
                break
            else:
                if hdr in ["Access-Control-Allow-Headers"]:
                    ownHeaders = set([h.strip() for h in self.headers[hdr].split(',')])
                    otherHeaders = set([h.strip() for h in other.headers[hdr].split(',')])
                    if ownHeaders != otherHeaders:
                        print("{} != {}".format(ownHeaders, otherHeaders))
                        headers_match = False
                        break
                else:
                    if self.headers[hdr] != other.headers[hdr]:
                        print("{} != {}".format(self.headers[hdr], other.headers[hdr]))
                        headers_match = False
                        break

        for (hdr, val) in other.headers:
            if hdr not in self.headers.keys():
                headers_match = False
                break

        bodies_match = (self.get_data() == other.get_data())
        if not bodies_match and self.mimetype == 'application/json' and other.mimetype == 'application/json':
            try:
                bodies_match = (json.loads(self.get_data()) == json.loads(other.get_data()))
            except ValueError:
                bodies_match = False

        return (bodies_match and
                (self.status == other.status) and
                headers_match and
                (self.mimetype == other.mimetype) and
                (self.content_type == other.content_type) and
                (self.direct_passthrough == other.direct_passthrough))


def proxied_request(uri, headers=None, data=None, method=None, proxies=None):
    uri, headers, data, method = flask_oauthlib.client.prepare_request(uri, headers, data, method)
    proxy = http.ProxyHandler(proxies)
    opener = http.build_opener(proxy)
    http.install_opener(opener)
    flask_oauthlib.client.log.debug('Request %r with %r method' % (uri, method))
    req = http.Request(uri, headers=headers, data=data)
    req.get_method = lambda: method.upper()
    try:
        resp = http.urlopen(req)
        content = resp.read()
        resp.close()
        return resp, content
    except http.HTTPError as resp:
        content = resp.read()
        resp.close()
        return resp, content


class WebAPI(object):
    def __init__(self, oauth_config=None):
        self.app = Flask(__name__)
        self.app.response_class = IppResponse
        self.app.before_first_request(self.torun)
        self.sockets = Sockets(self.app)
        self.socks = dict()
        self.port = 0

        # authentication/authorisation
        self._oauth_config = oauth_config
        self._authorize = self.default_authorize
        self._authenticate = self.default_authenticate

        self.add_routes(self, basepath='')

        # If the `fix_proxy` option is "enabled", interpret it as "use one proxy"
        if _config.get('fix_proxy') == 'enabled':
            if HAS_WERKZEUG_MIDDLEWARE:
                self.app.wsgi_app = ProxyFix(self.app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_port=1, x_prefix=1)
            else:
                self.app.wsgi_app = ProxyFix(self.app.wsgi_app)
        # If the `fix_proxy` option is a number, interpret it as the number of proxies to trust (Werkzeug>=0.15.0)
        elif HAS_WERKZEUG_MIDDLEWARE:
            try:
                proxy_count = int(_config.get('fix_proxy'))
                if proxy_count > 0:
                    self.app.wsgi_app = ProxyFix(
                        self.app.wsgi_app,
                        x_for=proxy_count,
                        x_proto=proxy_count,
                        x_host=proxy_count,
                        x_port=proxy_count,
                        x_prefix=proxy_count)
            except ValueError:
                # The `fix_proxy` option wasn't a number, so leave it off
                pass

    def add_routes(self, routesObject, basepath):

        assert not basepath.endswith('/'), "basepath must not end with a slash"

        def dummy(f):
            @wraps(f)
            def inner(*args, **kwargs):
                return f(*args, **kwargs)
            return inner

        def getbases(cls):
            bases = list(cls.__bases__)
            for x in cls.__bases__:
                bases += getbases(x)
            return bases

        # Populate a list of API paths to be used later for disambiguation
        api_paths = []
        for cls in [routesObject.__class__, ] + getbases(routesObject.__class__):
            for attr in cls.__dict__.keys():
                routesMethod = getattr(routesObject, attr)
                if hasattr(routesMethod, "common_path"):
                    api_paths.append(routesMethod.common_path)

        for cls in [routesObject.__class__, ] + getbases(routesObject.__class__):
            for attr in cls.__dict__.keys():
                routesMethod = getattr(routesObject, attr)
                if callable(routesMethod):
                    endpoint = "{}_{}".format(basepath.replace('/', '_'), routesMethod.__name__)
                    if hasattr(routesMethod, 'app_methods') and routesMethod.app_methods is not None:
                        methods = routesMethod.app_methods
                    else:
                        methods = ["GET", "HEAD"]
                    if hasattr(routesMethod, 'app_headers') and routesMethod.app_headers is not None:
                        headers = routesMethod.app_headers
                    else:
                        headers = []

                    crossdomain_methods = []
                    if hasattr(routesMethod, "common_path"):
                        stripped_path = routesMethod.common_path.rstrip("/")
                        trailing_slash = routesMethod.common_path.endswith("/")
                        if trailing_slash and stripped_path not in api_paths or not trailing_slash:
                            crossdomain_methods = ["OPTIONS", ]

                    if hasattr(routesMethod, "secure_route"):
                        self.app.route(
                            basepath + routesMethod.secure_route,
                            endpoint=endpoint,
                            methods=methods + crossdomain_methods)(
                                crossdomain(
                                    origin=routesMethod.app_origin,
                                    methods=methods,
                                    headers=headers + ['Content-Type', 'Authorization', 'token', ])(
                                        returns_requires_auth(routesMethod)))

                    elif hasattr(routesMethod, "response_route"):
                        self.app.route(
                            basepath + routesMethod.response_route,
                            endpoint=endpoint,
                            methods=["GET", "POST", "HEAD"] + crossdomain_methods)(
                                crossdomain(
                                    origin='*',
                                    methods=['GET', 'POST', 'HEAD'],
                                    headers=['Content-Type', 'Authorization', ])(
                                        returns_response(routesMethod)))

                    elif hasattr(routesMethod, "app_route"):
                        if routesMethod.app_auto_json:
                            self.app.route(
                                basepath + routesMethod.app_route,
                                endpoint=endpoint,
                                methods=methods + crossdomain_methods)(
                                    crossdomain(
                                        origin=routesMethod.app_origin,
                                        methods=methods,
                                        headers=headers + ['Content-Type', 'Authorization', ])(
                                            returns_json(routesMethod)))
                        else:
                            self.app.route(
                                basepath + routesMethod.app_route,
                                endpoint=endpoint,
                                methods=methods + crossdomain_methods)(
                                    crossdomain(
                                        origin=routesMethod.app_origin,
                                        methods=methods,
                                        headers=headers + ['Content-Type', 'Authorization', ])(
                                            dummy(routesMethod)))

                    elif hasattr(routesMethod, "app_file_route"):
                        self.app.route(
                            basepath + routesMethod.app_file_route,
                            endpoint=endpoint,
                            methods=methods + crossdomain_methods)(
                                crossdomain(
                                    origin='*',
                                    methods=methods,
                                    headers=headers + ['Content-Type', 'Authorization', ])(
                                        returns_file(routesMethod)))

                    elif hasattr(routesMethod, "app_resource_route"):
                        f = crossdomain(origin='*',
                                        methods=methods,
                                        headers=headers + ['Content-Type', 'Authorization', 'api-key', ])(
                                            returns_json(obj_path_access(routesMethod)))
                        self.app.route(
                            basepath + routesMethod.app_resource_route,
                            methods=methods + crossdomain_methods,
                            endpoint=endpoint)(f)
                        f.__name__ = endpoint + '_path'
                        self.app.route(
                            basepath + routesMethod.app_resource_route + '<path:path>/',
                            methods=methods + crossdomain_methods,
                            endpoint=f.__name__)(f)

                    elif hasattr(routesMethod, "socket_path"):
                        websocket_opened = getattr(routesObject, "on_websocket_connect", None)
                        if websocket_opened is None:
                            self.sockets.route(
                                basepath + routesMethod.socket_path,
                                endpoint=endpoint)(
                                    self.handle_sock(
                                        expects_json(routesMethod), self.socks))
                        else:
                            self.sockets.route(
                                basepath + routesMethod.socket_path,
                                endpoint=endpoint)(
                                    websocket_opened(
                                        expects_json(routesMethod)))

                    elif hasattr(routesMethod, "errorhandler_args"):
                        if routesMethod.errorhandler_args:
                            self.app.errorhandler(
                                *routesMethod.errorhandler_args,
                                **routesMethod.errorhandler_kwargs)(
                                    crossdomain(
                                        origin='*',
                                        methods=["GET", "POST", "PUT", "DELETE", "OPTIONS", "HEAD"])(
                                            dummy(routesMethod)))
                        else:
                            for n in range(400, 600):
                                try:  # Apply errorhandlers for all known error codes
                                    self.app.errorhandler(n)(crossdomain(
                                        origin='*',
                                        methods=["GET", "POST", "PUT", "DELETE", "OPTIONS", "HEAD"])(
                                            dummy(routesMethod)))
                                except KeyError:
                                    # Some error codes aren't valid
                                    pass

    @errorhandler()
    def error(self, e):
        if request.method == 'HEAD':
            if isinstance(e, HTTPException):
                return IppResponse('', status=e.code)
            else:
                return IppResponse('', 500)

        bm = request.accept_mimetypes.best_match(['application/json', 'text/html'])
        (exceptionType, exceptionParam, trace) = sys.exc_info()
        if bm == 'text/html':
            if isinstance(e, HTTPException):
                if e.code == 400:
                    status_code = e.code
                    response_title = "{}: {}".format(e.code, e.description)
                else:
                    return e.get_response()
            else:
                status_code = 500
                response_title = '500: Internal Exception'

            return IppResponse(
                highlight(
                    '\n'.join(traceback.format_exception(exceptionType, exceptionParam, trace)),
                    PythonTracebackLexer(),
                    HtmlFormatter(
                        linenos='table',
                        full=True,
                        title=response_title
                    )
                ),
                status=status_code,
                mimetype='text/html'
            )
        else:
            headers = e.headers if hasattr(e, 'headers') else None
            if isinstance(e, HTTPException):
                status_code = e.code
                error_description = e.description
            elif isinstance(e, AuthlibHTTPError):
                status_code = e.status_code
                error_description = e.get_error_description() if e.get_error_description() else e.error
            elif isinstance(e, AuthlibBaseError):
                status_code = 400
                error_description = e.message if e.description else e.error  # message = error : description
            else:
                status_code = 500
                error_description = 'Internal Error'

            response = {
                'code': status_code,
                'error': error_description,
                'debug': str({
                    'traceback': [str(x) for x in traceback.extract_tb(trace)],
                    'exception': [str(x) for x in traceback.format_exception_only(exceptionType, exceptionParam)]
                })
            }
            return IppResponse(json.dumps(response), status=status_code, mimetype='application/json', headers=headers)

    def torun(self):  # pragma: no cover
        pass

    def stop(self):  # pragma: no cover
        pass

    def handle_sock(self, func, socks):
        @wraps(func)
        def inner_func(ws, **kwds):
            sock_uuid = uuid.uuid4()
            socks[sock_uuid] = ws
            print("Opening Websocket {} at path /, Receiving ...".format(sock_uuid))
            while True:
                try:
                    message = ws.receive()
                except Exception:
                    message = None

                if message is not None:
                    func(ws, message, **kwds)
                    continue
                else:
                    print("Websocket {} closed".format(sock_uuid))
                    del socks[sock_uuid]
                    break
        return inner_func

    def default_authorize(self, token):
        if self._oauth_config is not None:
            # Ensure the user is permitted to use function
            print('authorizing: {}'.format(token))
            loginserver = self._oauth_config['loginserver']
            proxies = self._oauth_config['proxies']
            whitelist = self._oauth_config['access_whitelist']
            result = proxied_request(
                        uri="{}/check-token".format(loginserver),
                        headers={'token': token}, proxies=proxies)
            json_payload = json.loads(result[1])
            return json_payload['userid'] in whitelist
        else:
            return True

    def default_authenticate(self, token):
        if self._oauth_config is not None:
            # Validate the token that the webapp sends
            loginserver = self._oauth_config['loginserver']
            proxies = self._oauth_config['proxies']
            print('authenticating: {}'.format(token))
            if token is None:
                return False
            result = proxied_request(
                        uri="{}/check-token".format(loginserver),
                        headers={'token': token}, proxies=proxies)
            print('token result code: {}'.format(result[0].code))
            print('result payload: {}'.format(result[1]))
            return result[0].code == 200 and json.loads(result[1])['token'] == token
        else:
            return True

    def authorize(self, f):
        self._authorize = f
        return f

    def authenticate(self, f):
        self._authenticate = f
        return f



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain


class UnparseError(Exception):
    pass


def flatten(xs):
    for x in xs:
        if isinstance(x, list):
            for y in flatten(x):
                yield y
        else:
            yield x


OPERATOR_MAP = {
    'out': 'nin',
    'contains': 'all'
}


TYPE_A = '$eq'.split()  # $eq is not an operator in mongodb. Type A handles implicit.
TYPE_B = '$lte $lt $gt $gte $ne'.split()
TYPE_C = '$in $nin $all'.split()
TYPE_D = '$or $nor'.split()
TYPE_E = '$not'.split()
TYPE_F = '$and'.split()


def handle_type_a(_, field, value):
    return {field: value}


def handle_type_b(operator, field, value):
    return {field: {operator: value}}


def handle_type_c(operator, field, *values):
    return {field: {operator: list(flatten(values))}}


def handle_type_d(operator, *values):
    return {operator: list(values)}


def handle_type_e(operator, query):
    return {key: {operator: value} for key, value in query.items()}


def handle_type_f(operator, *values):
    ''' handles implicit collapsing of some operators:
        eg {$and: [a, b]} -> {a, b}
        {$and: [a]} -> a
        {$and: [a->b,a->c]} -> {$and: [a->b,a->c]}
    '''
    all_values_are_dicts = all(isinstance(val, dict) for val in values)
    if all_values_are_dicts:
        no_duplicate_keys = len(set(chain.from_iterable(d.keys() for d in values))) == len(values)

    if all_values_are_dicts and no_duplicate_keys:
        return dict(chain.from_iterable(d.items() for d in values))
    elif len(values) == 1:
        return values[0]
    else:
        return handle_type_d(operator, *values)


assemblers = dict(
    chain(
        ((operator, handle_type_a) for operator in TYPE_A),
        ((operator, handle_type_b) for operator in TYPE_B),
        ((operator, handle_type_c) for operator in TYPE_C),
        ((operator, handle_type_d) for operator in TYPE_D),
        ((operator, handle_type_e) for operator in TYPE_E),
        ((operator, handle_type_f) for operator in TYPE_F)
    )
)


def translate_name(name):
    return '$' + OPERATOR_MAP.get(name, name)


def unparse(node):
    if not isinstance(node, dict):
        return node

    try:
        name = translate_name(node['name'])
        args = [unparse(a) for a in node['args']]
    except KeyError:
        raise UnparseError('Abstract Sytnax Tree not in expected form: {}'.format(node))

    try:
        assembler = assemblers[name]
    except KeyError:
        raise UnparseError('Unsupported operator: {}'.format(name))

    return assembler(name, *args)



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import json
from os import path, sep
from functools import wraps
from flask import request
from OpenSSL import crypto
from six.moves.urllib.parse import urljoin, parse_qs
from requests.exceptions import RequestException

from nmoscommon.mdnsbridge import IppmDNSBridge
from nmoscommon.nmoscommonconfig import config as _config
from nmoscommon.logger import Logger as defaultLogger
from authlib.jose import jwt
from authlib.oauth2.rfc6749.errors import MissingAuthorizationError, \
    UnsupportedTokenTypeError
from authlib.common.errors import AuthlibBaseError

from .claims_options import IS_XX_CLAIMS
from .claims_validator import JWTClaimsValidator

MDNS_SERVICE_TYPE = "nmos-auth"
OAUTH_MODE = _config.get('oauth_mode', True)

NMOSAUTH_DIR = path.abspath(path.join(sep, 'var', 'nmosauth'))
CERT_FILE = 'certificate.pem'
CERT_FILE_PATH = path.join(NMOSAUTH_DIR, CERT_FILE)

APINAMESPACE = "x-nmos"
APINAME = "auth"
APIVERSION = "v1.0"
CERT_ENDPOINT = 'certs'
CERT_URL_PATH = '/{}/{}/{}/{}'.format(APINAMESPACE, APINAME, APIVERSION, CERT_ENDPOINT)


class RequiresAuth(object):

    def __init__(self, condition=OAUTH_MODE, claimsOptions=IS_XX_CLAIMS,
                 certificate=None):
        self.condition = condition
        self.claimsOptions = claimsOptions
        self.certificate = certificate
        self.bridge = IppmDNSBridge()
        self.logger = defaultLogger("authresource")

    def getHrefFromService(self, serviceType):
        return self.bridge.getHref(serviceType)

    def getCertFromEndpoint(self):
        try:
            href = self.getHrefFromService(MDNS_SERVICE_TYPE)
            certHref = urljoin(href, CERT_URL_PATH)
            self.logger.writeInfo('cert href is: {}'.format(certHref))
            cert_resp = requests.get(certHref, timeout=0.5, proxies={'http': ''})
            cert_resp.raise_for_status()  # Raise error if status !=200
        except RequestException as e:
            self.logger.writeError("Error: {0!s}".format(e))
            self.logger.writeError("Cannot find certificate at {}. Is the Auth Server Running?".format(certHref))
            raise

        contentType = cert_resp.headers['content-type'].split(";")[0]
        if contentType == "application/json":
            cert = cert_resp.json()
            try:
                if len(cert) > 1:
                    self.logger.writeWarning("Multiple certificates at Endpoint. Returning First Instance.")
                cert = cert[0]
                return cert
            except Exception as e:
                self.logger.writeError("Error: {}. Endpoint contains: {}".format(str(e), cert))
                raise
        else:
            self.logger.writeError("Incorrect Content-Type. Expected 'application/json but got {}".format(contentType))
            raise ValueError

    def getCertFromFile(self, filename):
        try:
            if filename is not None:
                with open(filename, 'r') as myfile:
                    cert_data = myfile.read()
                    self.certificate = cert_data
                    return cert_data
        except OSError:
            self.logger.writeError("File does not exist or you do not have permission to open it")
            raise

    def extractPublicKey(self, certificate):
        crtObj = crypto.load_certificate(crypto.FILETYPE_PEM, certificate)
        pubKeyObject = crtObj.get_pubkey()
        pubKeyString = crypto.dump_publickey(crypto.FILETYPE_PEM, pubKeyObject)
        if pubKeyString is None:
            self.logger.writeError("Public Key could not be extracted from certificate")
            raise ValueError
        else:
            return pubKeyString.decode('utf-8')

    def getPublicKey(self):
        if self.certificate is None:
            self.logger.writeInfo("Fetching Certificate...")
            try:
                self.logger.writeInfo("Trying to fetch cert using mDNS...")
                cert = self.getCertFromEndpoint()
            except Exception as e:
                self.logger.writeError("Error: {0!s}. Trying to fetch Cert From File...".format(e))
                cert = self.getCertFromFile(CERT_FILE_PATH)
            self.certificate = cert
        pubKey = self.extractPublicKey(self.certificate)
        return pubKey

    def processAccessToken(self, auth_string):
        # Auth string is of type 'Bearer xAgy65..'
        if auth_string.find(' ') > -1:
            token_type, token_string = auth_string.split(None, 1)
            if token_type.lower() != "bearer":
                raise UnsupportedTokenTypeError()
        # Otherwise string is access token 'xAgy65..'
        else:
            token_string = auth_string
        if token_string == "null" or token_string == "":
            raise MissingAuthorizationError()
        pubKey = self.getPublicKey()
        claims = jwt.decode(s=token_string, key=pubKey,
                            claims_cls=JWTClaimsValidator,
                            claims_options=self.claimsOptions,
                            claims_params=None)
        claims.validate()

    def handleHttpAuth(self):
        """Handle bearer token string ("Bearer xAgy65...") in "Authorzation" Request Header"""
        auth_string = request.headers.get('Authorization', None)
        if auth_string is None:
            raise MissingAuthorizationError()
        self.processAccessToken(auth_string)

    def handleSocketAuth(self, *args, **kwargs):
        """Handle bearer token string ("access_token=xAgy65...") in Websocket URL Query Param"""
        ws = args[0]
        environment = ws.environ
        auth_header = environment.get('HTTP_AUTHORIZATION', None)
        if auth_header is not None:
            self.logger.writeInfo("auth header string is {}".format(auth_header))
            auth_string = auth_header
        else:
            self.logger.writeWarning("Websocket does not have auth header, looking in query string..")
            query_string = environment.get('QUERY_STRING', None)
            try:
                if query_string is not None:
                    try:
                        auth_string = parse_qs(query_string)['access_token'][0]
                    except KeyError:
                        self.logger.writeError("""
                            'access_token' URL param doesn't exist. Websocket authentication failed.
                        """)
                        raise MissingAuthorizationError()
                self.processAccessToken(auth_string)
            except AuthlibBaseError:
                err = {"type": "error", "data": "Socket Authentication Error"}
                ws.send(json.dumps(err))
                raise

    def JWTRequired(self):
        def JWTDecorator(func):
            @wraps(func)
            def processAccessToken(*args, **kwargs):
                # Check to see if request is a Websocket upgrade, else treat request as a standard HTTP request
                headers = request.headers
                if ('Upgrade' in headers.keys() and headers['Upgrade'].lower() == 'websocket'):
                    self.logger.writeInfo("Using Socket handler")
                    self.handleSocketAuth(*args, **kwargs)
                else:
                    self.logger.writeInfo("Using HTTP handler")
                    self.handleHttpAuth()
                return func(*args, **kwargs)
            return processAccessToken
        return JWTDecorator

    def __call__(self, func):
        if not self.condition:
            # Return the function unchanged, not decorated.
            return func

        # Return decorated function
        decorator = self.JWTRequired()
        return decorator(func)



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zeroconf_monkey as zeroconf
from nmoscommon.logger import Logger
from nmoscommon.nmoscommonconfig import config
from .dnsServiceController import DNSServiceController
from .mdnsListener import MDNSListener
from .mdnsCallbackHandler import MDNSAdvertisementCallbackHandler
from .mdnsRegistration import MDNSRegistration
from .mdnsExceptions import ServiceAlreadyExistsException, InterfaceNotFoundException, ServiceNotFoundException
from .mdnsInterfaceController import MDNSInterfaceController
from .mdnsRegistrationController import MDNSRegistrationController
from .mdnsSubscriptionController import MDNSSubscriptionController


class MDNSEngine(object):

    def __init__(self):
        self.running = False
        self.dnsServiceControllers = []

    def start(self):
        if not self.running:
            self.logger = Logger('mdns-engine')
            self.subscriptionController = MDNSSubscriptionController()
            self.registrationController = MDNSRegistrationController()
            self.interfaceController = MDNSInterfaceController(self.logger)
            self.running = True

    def stop(self):
        self.close()

    def close(self):
        self.logger.writeDebug("MDNS Engine Closed")
        self.subscriptionController.close()
        self.registrationController.close()
        self.interfaceController.close()
        for controller in self.dnsServiceControllers:
            controller.close()
        self.running = False

    def run(self):
        pass

    def _autostart_if_required(self):
        if not self.running:
            self.start()

    def register(self, name, regtype, port, txtRecord="", callback=None, interfaceIps=None):
        self._autostart_if_required()
        callbackHandler = MDNSAdvertisementCallbackHandler(callback, regtype, name, port, txtRecord)
        if not interfaceIps:
            interfaceIps = []
        try:
            interfaces = self.interfaceController.getInterfaces(interfaceIps)
        except InterfaceNotFoundException:
            callbackHandler.entryFailed()
            return
        registration = MDNSRegistration(interfaces, name, regtype, port, txtRecord)
        self._add_registration_handle_errors(registration, callbackHandler)

    def _add_registration_handle_errors(self, registration, callbackHandler):
        try:
            self.registrationController.addRegistration(registration)
        except zeroconf.NonUniqueNameException:
            callbackHandler.entryCollision()
        except ServiceAlreadyExistsException:
            callbackHandler.entryCollision()
        except zeroconf.Error as e:
            callbackHandler.entryFailed()
            print(str(e))
        else:
            callbackHandler.entryEstablished()

    def update(self, name, regtype, txtRecord=None):
        self._autostart_if_required()
        try:
            registration = self.registrationController.registrations[regtype][name]
        except KeyError:
            self.logger.writeError("Could not update registration type: {} with name {}"
                                   " - registration not found".format(regtype, name))
            raise ServiceNotFoundException
        registration.update(name=name, regtype=regtype, txtRecord=txtRecord)

    def unregister(self, name, regType):
        self._autostart_if_required()
        self.registrationController.removeRegistration(name, regType)

    def callback_on_services(self, regtype, callback, registerOnly=True, domain=None):
        self._autostart_if_required()
        doDNSDiscover = config['dns_discover']
        domDNSDiscover = config['mdns_discover']
        if domDNSDiscover:
            listener = MDNSListener(callback, registerOnly)
            self.subscriptionController.addSubscription(listener, regtype)
        if doDNSDiscover:
            dnsServiceController = DNSServiceController(
                regtype,
                callback,
                self.logger,
                registerOnly,
                domain
            )
            dnsServiceController.start()
            self.dnsServiceControllers.append(dnsServiceController)


if __name__ == "__main__":
    __package__ = "nmoscommon.mdns.zeroconfEngine"

    def callback(callbackObject):
        print("----------------")
        print(callbackObject)
        # pass
    engine = MDNSEngine()
    engine.start()
    engine.callback_on_services("_nmos-node._tcp", callback, domain=None)
    """engine.register(
        "registration_http",
        "_nmos-registration._tcp",
        8080,
        {
            'pri': 200,
            'api_ver': 'v1.0',
            'api_proto': 'https'
        },
        callback,
        ["172.29.82.49", "172.29.80.118", "127.0.0.1"]
    )"""
    try:
        input("Press enter to update registration...\n\n")
    except Exception:
        pass
    finally:
        # engine.update("registration_http", "_nmos-registration._tcp", {"test": "text"})
        try:
            input("Press enter to exit...\n\n")
        except Exception:
            pass
        finally:
            engine.stop()



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dns.resolver  # Importing dns alone is not enough for NXDOMAIN to work
from nmoscommon.mdns.mdnsExceptions import DNSRecordNotFound, NoDefaultDnsSearchDomain


def _defaultDomain():
    try:
        resolve = dns.resolver.Resolver()
        return str(resolve.search[0])
    except IndexError:
        raise NoDefaultDnsSearchDomain


def _appendDomain(record, domainName):
    record = record + "." + domainName
    return record


def _dnsRequest(record, recordType, addDomain=True, customDomain=None):
    if addDomain:
        if customDomain:
            record = _appendDomain(record, customDomain)
        else:
            record = _appendDomain(record, _defaultDomain())
    try:
        answers = dns.resolver.query(record, recordType)
    except dns.resolver.NXDOMAIN:
        print("Could not find {} type record at {}".format(recordType, record))
        raise DNSRecordNotFound
    else:
        return answers


def checkDNSSDActive(customDomain=None):
    try:
        _dnsRequest('lb._dns-sd._udp', 'PTR', True, customDomain)
    except DNSRecordNotFound:
        return False
    else:
        return True


def getServiceTypes(customDomain=None):
    return _dnsRequest('_services._dns-sd._udp', 'PTR', True, customDomain)


def discoverService(regType, customDomain=None):
    return _dnsRequest(regType, 'PTR', True, customDomain)


def getTXTRecord(service):
    return _dnsRequest(service, 'TXT', False)


def getSRVRecord(service):
    return _dnsRequest(service, 'SRV', False)



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
import zmq.green as zmq
import uuid
import os
import os.path
import gevent
import traceback
import json
import stat
import warnings


class RemoteException(Exception):
    pass


class LocalException(Exception):
    pass


# Included for backward compatibility with previous typo'ed version
RemoteExcepton = RemoteException


class Host(object):
    """This class provides a server which can make a set of ipc commands available at a well known address.
    It provides a decorator, @ipcmethod which is used to decorate methods that are to be callable remotely.

    The server itself should be started with start and stopped with stop when no longer needed, though exiting
    the main application thread will also shut it down, this is not clean and not recommended."""
    def __init__(self, address, timeout=100):
        self.address = address
        ctx = zmq.Context.instance()

        self.timeout = timeout

        self.socket = ctx.socket(zmq.REP)
        self.socket.bind(self.address)
        self.socket.setsockopt(zmq.LINGER, 0)

        if address[:6] == "ipc://":
            os.chmod(address[6:], (stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
                                   stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP |
                                   stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH))

        self.thread = None
        self._stop = True

        self.methods = {}

        self.methods['getmethods'] = self.getmethods
        self.greenlet = None

    def start(self):
        if self._stop:
            self._stop = False
            self.greenlet = gevent.spawn(self._run)

    def stop(self):
        self._stop = True
        if self.greenlet is not None:
            self.greenlet.kill()
        self.greenlet = None
        if not self.socket.closed:
            self.socket.close()

    def _run(self):
        while not self._stop:
            r = self.socket.poll(timeout=self.timeout)
            if r != 0:
                msg = self.socket.recv_json()
                if ('function' not in msg or 'args' not in msg or 'kwargs' not in msg):
                    self.socket.send_json({})
                    continue

                if msg['function'] not in self.methods:
                    self.socket.send_json({'exc': 'AttributeError'})
                    continue

                try:
                    r = self.methods[msg['function']](*(msg['args']), **(msg['kwargs']))
                except Exception:
                    self.socket.send_json({'exc': traceback.format_exc()})
                    continue
                if r is not None:
                    self.socket.send_json({'ret': r})
                else:
                    self.socket.send_json({})

    def ipcmethod(self, name=None):
        def _inner(function):
            _name = name
            if _name is None:
                _name = function.__name__
            self.methods[_name] = function
            return function
        return _inner

    def getmethods(self):
        """Return a list of all methods available and their help text"""
        return dict(
            ((key, self.methods[key].__doc__ if self.methods[key].__doc__ is not None else "") for key in self.methods.keys())
        )


class Proxy(object):
    """This class provides a proxy to a remote ipc host which can be used to invoke methods on that host."""
    def __init__(self, address, timeout=100):
        self.address = address
        ctx = zmq.Context.instance()

        self.timeout = timeout

        if address[:6] == "ipc://":
            if not os.path.exists(address[6:]):
                raise RuntimeError

        self.socket = ctx.socket(zmq.REQ)
        self.socket.connect(self.address)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.setsockopt(zmq.SNDTIMEO, 0)
        self.socket.setsockopt(zmq.RCVTIMEO, 0)

    def invoke_named(self, name, *args, **kwargs):
        try:
            msg = {'function': name,
                   'args': args,
                   'kwargs': kwargs}
            gevent.sleep(0)
            self.socket.send_json(msg)
            gevent.sleep(0)

            if self.socket.poll(timeout=self.timeout) == 0:
                self.close()
                raise LocalException("Unconnected Socket")

            r = self.socket.recv_json()
            gevent.sleep(0)
            if 'exc' in r:
                raise RemoteException(r['exc'])
            if 'ret' in r:
                return r['ret']
        except Exception:
            gevent.sleep(0)
            raise

    def close(self):
        if self.socket:
            try:
                self.socket.close()
            except Exception:
                pass
            self.socket = None

    def __getattr__(self, name):
        def _invoke(*args, **kwargs):
            return self.invoke_named(name, *args, **kwargs)
        return _invoke


# This is deprecated
class Socket(object):  # pragma: no cover
    def __init__(self, name=None, rmethods=None):
        warnings.warn("The Socket Class is Deprecated and will be removed in future", DeprecationWarning, stacklevel=1)
        if rmethods is None:
            rmethods = []
        if name is None:
            name = '/tmp/nmoscommon_ipc_' + str(uuid.uuid4())
            self.side = "slave"
        else:
            self.side = "master"

        self.name = name
        ctx = zmq.Context.instance()
        if self.side == "master":
            self.socket = ctx.socket(zmq.REQ)
            self.socket.connect("ipc://" + self.name)
        else:
            self.socket = ctx.socket(zmq.REP)
            self.socket.bind("ipc://" + self.name)
        self.socket.setsockopt(zmq.LINGER, 0)

        self.greenlet = None
        self._stop = True

        self.lmethods = dict()
        self.rmethods = []
        for rmethod in rmethods:
            self.remote(rmethod)

    def start(self):
        if self.side == "slave":
            if self._stop:
                self._stop = False
                self.greenlet = gevent.spawn(self._run)

    def stop(self):
        self._stop = True
        if self.greenlet is not None:
            self.greenlet.kill()
        self.greenlet = None

    def _run(self):
        while not self._stop:
            r = self.socket.poll(timeout=100)
            if r != 0:
                msg = self.socket.recv_json()
                if msg["function"] not in self.lmethods:
                    rmsg = None
                else:
                    rmsg = self.lmethods[msg["function"]](*msg["args"], **msg["kwargs"])
                self.socket.send_json({"return": rmsg})

    def request(self, function, *args, **kwargs):
        self.socket.send_json({
            "function": function,
            "args": args,
            "kwargs": kwargs})
        self.socket.poll(timeout=100)
        rmsg = self.socket.recv_json()
        if "return" in rmsg:
            return rmsg["return"]

    def local(self, f):
        self.lmethods[f.__name__] = f
        return f

    def remote(self, funcname):
        self.rmethods.append(funcname)

    def __getattr__(self, name):
        def callremote(*args, **kwargs):
            return self.request(name, *args, **kwargs)

        try:
            return super(Socket, self).__getattr__(name)
        except AttributeError:
            if name in self.rmethods:
                return callremote
            raise


def main():
    import sys
    address = "ipc:///tmp/nmoscommon_test"
    if len(sys.argv) > 1:
        address = sys.argv[1]

    if len(sys.argv) > 2:
        p = Proxy(address)
        print(json.dumps(getattr(p, sys.argv[2])(*(sys.argv[3:]))))
    else:
        h = Host(address)

        @h.ipcmethod("hello")
        def hello(name="James"):
            return "Hello, {}".format(name)
        print("Listening on {}".format(address))
        h._stop = False
        h._run()


if __name__ == "__main__":  # pragma: no cover
    main()



IS_XX_CLAIMS = {
    "iat": {"essential": True},
    "nbf": {"essential": False},
    "exp": {"essential": True},
    "iss": {"essential": True},
    "sub": {"essential": True},
    "aud": {"essential": True},
    "scope": {"essential": True},
    "x-nmos-api": {"essential": True}
}

IS_04_REG_CLAIMS = {
    "iat": {"essential": True},
    "nbf": {"essential": False},
    "exp": {"essential": True},
    "iss": {"essential": True},
    "sub": {"essential": True},
    "aud": {"essential": True},
    "scope": {
        "essential": True,
        "value": "is-04"
    },
    "x-nmos-api": {
        "essential": True,
        "value": {
            "name": "is-04",
            "access": "write"
        }
    }
}

IS_04_QUERY_CLAIMS = {
    "iat": {"essential": True},
    "nbf": {"essential": False},
    "exp": {"essential": True},
    "iss": {"essential": True},
    "sub": {"essential": True},
    "aud": {"essential": True},
    "scope": {
        "essential": True,
        "value": "is-04"
    },
    "x-nmos-api": {
        "essential": True,
        "value": {
            "name": "is-04",
            "access": "write"
        }
    }
}

IS_04_NODE_CLAIMS = {
    "iat": {"essential": True},
    "nbf": {"essential": False},
    "exp": {"essential": True},
    "iss": {"essential": True},
    "sub": {"essential": True},
    "aud": {"essential": True},
    "scope": {
        "essential": True,
        "value": "is-04"
    },
    "x-nmos-api": {
        "essential": True,
        "value": {
            "name": "is-04",
            "access": "write"
        }
    }
}

IS_05_CLAIMS = {
    "iat": {"essential": True},
    "nbf": {"essential": False},
    "exp": {"essential": True},
    "iss": {"essential": True},
    "sub": {"essential": True},
    "aud": {"essential": True},
    "scope": {
        "essential": True,
        "value": "is-05"
    },
    "x-nmos-api": {
        "essential": True,
        "value": {
            "name": "is-05",
            "access": "write"
        }
    }
}



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DNSListener(object):

    def __init__(self, callback, registerOnly):
        self.registerOnly = registerOnly
        self.callback = callback

    def addListener(self, subscription):
        callbackData = self._prepareCallback("add", subscription)
        self.callback(callbackData)

    def removeListener(self, subscription):
        if not self.registerOnly:
            callbackData = self._prepareCallback("remove", subscription)
            self.callback(callbackData)

    def _prepareCallback(self, action, subscription):
        toReturn = {}
        toReturn['action'] = action
        toReturn['address'] = subscription.address
        toReturn['port'] = subscription.port
        toReturn['type'] = subscription.type
        toReturn['txt'] = subscription.txt
        toReturn['name'] = subscription.name
        toReturn['hostname'] = subscription.hostname
        return toReturn



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function

import uuid
import json
import os
import copy
import netifaces

from .logger import Logger

logger = Logger("utils", None)

DOWNGRADE_MAP = {
    "v1.1": {
        "node": ["description", "tags", "api", "clocks"],
        "device": ["description", "tags", "controls"],
        "source": ["clock_name", "grain_rate", "channels"],
        "flow": [
            "device_id", "grain_rate", "media_type", "sample_rate", "bit_depth", "DID_SDID", "frame_width",
            "frame_height", "interlace_mode", "colorspace", "components", "transfer_characteristic"
        ],
    },
    "v1.2": {
        "node": ["interfaces"],
        "sender": ["interface_bindings", "caps", "subscription"],
        "receiver": ["interface_bindings"]
    },
    "v1.3": {
        "node": ["attached_network_device", "authorization"],
        "device": ["authorization"],
        "source": ["event_type"],
        "flow": ["event_type"],
    }
}


def get_node_id():
    node_id_path = "/var/nmos-node/facade.json"
    node_id = str(uuid.uuid1())
    try:
        if os.path.exists(node_id_path):
            f = open(node_id_path, "r")
            node_id = json.loads(f.read())["node_id"]
            f.close()
        else:
            f = open(node_id_path, "w")
            f.write(json.dumps({"node_id": node_id}))
            f.close()
    except Exception as e:
        logger.writeWarning("Unable to read or write node ID. Using dynamically generated ID")
        logger.writeWarning(str(e))
    return node_id


def getLocalIP():
    interfaces = netifaces.interfaces()
    for interface in interfaces:
        if (interface is not None) & (str(interface)[0:2] != 'lo'):
            try:
                for addr in netifaces.ifaddresses(interface)[netifaces.AF_INET]:
                    if str(addr['addr'])[0:4] != "127.":
                        return addr['addr']
            except KeyError:
                pass
    # Could not find an interface
    return None


def translate_api_version(
    obj, resource_type, target_ver, downgrade_ver=None, translation_map=None, remove_keys=["max_api_version"]
):
    if translation_map is None:
        translation_map = DOWNGRADE_MAP

    # Ensure rtype matches keys inside downgrade map
    rtype = resource_type.rstrip('s')

    # Sort version list in ascending order
    version_list = sorted(translation_map.keys())
    # Add 'v1.0' as first element in version list
    version_list.insert(0, 'v1.0')
    # Set the max version as last element in list
    max_ver = version_list[-1]

    # Create deep copy of object so as to not mutate original
    obj = copy.deepcopy(obj)

    # Get current API version (present in the query API implementation), or default to max API version if absent
    current_api_version = obj.get("@_apiversion", max_ver)

    # Fail if target API version is greater than maximum
    if api_ver_compare(target_ver, max_ver) > 0:
        return None

    # Downgrade API object, for a given resource, until it reaches target_version
    while api_ver_compare(target_ver, current_api_version) < 0:
        logger.writeDebug("Processing downgrading from {}".format(current_api_version))
        obj = remove_keys_from_resource(obj, rtype, translation_map[current_api_version])
        current_api_version = version_list[version_list.index(current_api_version) - 1]

    # Set api version key to new downgraded api version
    if "@_apiversion" in obj:
        obj["@_apiversion"] = current_api_version

    # Delete any implementation-specific keys not in the spec
    [obj.pop(key) for key in remove_keys if key in obj]

    # Check if the object's API version is permitted in the output
    if target_ver == current_api_version:
        return obj
    # Where `downgrade_ver` relates to the query.downgrade query parameter of the Query API
    elif downgrade_ver and api_ver_compare(current_api_version, downgrade_ver) >= 0:
        return obj

    # Fallback
    return None


def api_ver_compare(first, second):
    ver_first = first[1:].split(".")
    ver_second = second[1:].split(".")
    if ver_first[0] < ver_second[0]:
        return -1
    elif ver_first[0] > ver_second[0]:
        return 1
    elif ver_first[1] < ver_second[1]:
        return -2
    elif ver_first[1] > ver_second[1]:
        return 2
    else:
        return 0


def remove_keys_from_resource(obj, rtype, translation_map):
    if rtype in translation_map:
        for key in translation_map[rtype]:
            _remove_if_present(obj, key)
    return obj


def _remove_if_present(obj, delete_key):
    if isinstance(obj, list):
        for element in obj:
            _remove_if_present(element, delete_key)
    if isinstance(obj, dict):
        for key, val in obj.copy().items():
            if isinstance(val, dict):
                val = _remove_if_present(val, delete_key)
            if isinstance(val, list):
                for element in val:
                    val = _remove_if_present(element, delete_key)
            if key == delete_key:
                del obj[delete_key]
    return obj



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
import os
from .ipc import Proxy
import gevent
from threading import Lock
from nmoscommon.logger import Logger
from copy import deepcopy

FAC_SUCCESS = 0
FAC_EXISTS = 1
FAC_UNREGISTERED = 2
FAC_UNAUTHORISED = 3
FAC_UNSUPPORTED = 4
FAC_OTHERERROR = 5


class Facade(object):
    """This class serves as a proxy for the Facade running on the same machine if it exists. If no facade exists
    on this machine then it will do nothing, but calls will still function without throwing any exceptions."""
    def __init__(self, srv_type, address="ipc:///tmp/ips-nodefacade", logger=None):

        self.logger = Logger("facade_proxy", logger)
        self.logger.writeWarning("This class is deprecated. Please use the matching one in nmos-node instead.")
        self.ipc = None
        self.srv_registered = False  # Flag whether service is registered
        self.reregister = False  # Flag whether resources are correctly registered
        self.address = address
        self.srv_type = srv_type.lower()
        self.srv_type_urn = "urn:x-ipstudio:service:" + self.srv_type
        self.pid = os.getpid()
        self.resources = {}
        self.controls = {}
        self.href = None
        self.proxy_path = None
        self.lock = Lock()  # Protect access to IPC socket

    def setup_ipc(self):
        with self.lock:
            try:
                self.ipc = Proxy(self.address)
            except Exception:
                self.ipc = None

    def register_service(self, href, proxy_path):
        self.logger.writeInfo("Register service")
        self.href = href
        self.proxy_path = proxy_path
        if not self.ipc:
            self.setup_ipc()
        if not self.ipc:
            return
        try:
            with self.lock:
                s = self.ipc.srv_register(self.srv_type, self.srv_type_urn, self.pid, href, proxy_path)
                if s == FAC_SUCCESS:
                    self.srv_registered = True
                else:
                    self.logger.writeInfo("Service registration failed: {}".format(self.debug_message(s)))
        except Exception as e:
            self.logger.writeError("Exception when registering service: {}".format(str(e)))
            self.ipc = None

    def unregister_service(self):
        if not self.ipc:
            self.setup_ipc()
        if not self.ipc:
            return
        try:
            with self.lock:
                self.ipc.srv_unregister(self.srv_type, self.pid)
                self.srv_registered = False
        except Exception as e:
            self.logger.writeError("Exception when unregistering service: {}".format(str(e)))
            self.ipc = None

    def heartbeat_service(self):
        if not self.ipc:
            self.setup_ipc()
        if not self.ipc:
            return
        try:
            with self.lock:
                s = self.ipc.srv_heartbeat(self.srv_type, self.pid)
                if s != FAC_SUCCESS:
                    self.srv_registered = False
                    self.logger.writeInfo("Heartbeat failed: {}".format(self.debug_message(s)))
                else:
                    self.srv_registered = True
            if not self.srv_registered or self.reregister:
                # Handle reconnection if facade disappears
                self.logger.writeInfo("Reregistering all services")
                self.reregister_all()
        except Exception as e:
            self.logger.writeError("Exception when heartbeating service: {}".format(str(e)))
            self.ipc = None

    # ONLY call this directly from within heartbeat_service!
    # To cause a re-registration on failure, set self.reregister!
    def reregister_all(self):
        self.unregister_service()
        if self.srv_registered:
            return
        self.register_service(self.href, self.proxy_path)
        if not self.srv_registered:
            return

        # TODO(clyntp): the following blocks are so similar...

        # re-register resources
        for type in self.resources:
            for key in self.resources[type]:
                try:
                    with self.lock:
                        resource = self.resources[type][key]
                        # Hide some implementation details for receivers
                        if type == "receiver":
                            resource = deepcopy(self.resources[type][key])
                            if "pipel_id" in self.resources[type][key]:
                                resource.pop('pipel_id')
                            if "pipeline_id" in self.resources[type][key]:
                                resource.pop('pipeline_id')
                        self.ipc.res_register(self.srv_type, self.pid, type, key, resource)
                except Exception as e:
                    self.logger.writeError("Exception when re-registering resource: {}".format(str(e)))
                    self.ipc = None
                    gevent.sleep(0)
                    return

        # re-register controls
        for device_id in self.controls:
            for control_href in self.controls[device_id]:
                control_data = self.controls[device_id][control_href]
                try:
                    with self.lock:
                        self.ipc.control_register(self.srv_type, self.pid, device_id, control_data)
                except Exception as e:
                    self.logger.writeError("Exception when re-registering control: {}".format(str(e)))
                    self.ipc = None
                    gevent.sleep(0)
                    return

        self.reregister = False

    def _call_ipc_method(self, method, *args, **kwargs):
        if not self.srv_registered:
            # Don't attempt if not registered - will just hit many timeouts
            self.reregister = True
            return
        if not self.ipc:
            self.setup_ipc()
        if not self.ipc:
            self.reregister = True
            return
        try:
            with self.lock:
                return self.ipc.invoke_named(method, self.srv_type, self.pid, *args, **kwargs)
        except Exception as e:
            self.logger.writeError("Exception when calling IPC method: {}".format(str(e)))
            self.ipc = None
            self.reregister = True

    def addResource(self, type, key, value):
        value = deepcopy(value)
        if type not in self.resources:
            self.resources[type] = {}
        self.resources[type][key] = value
        self._call_ipc_method("res_register", type, key, value)

    def updateResource(self, type, key, value):
        value = deepcopy(value)
        if type not in self.resources:
            self.resources[type] = {}
        self.resources[type][key] = value
        self._call_ipc_method("res_update", type, key, value)

    def delResource(self, type, key):
        if type in self.resources:
            # Hack until adoption of flow instances (Ensure transports for flow are deleted)
            if type == "flow" and "transport" in self.resources:
                for transport in tuple(self.resources["transport"].keys()):
                    if key == self.resources["transport"][transport]["flow-id"]:
                        del self.resources["transport"][transport]
            if key in self.resources[type]:
                del self.resources[type][key]
        self._call_ipc_method("res_unregister", type, key)

    def addControl(self, device_id, control_data):
        if device_id not in self.controls:
            self.controls[device_id] = {}
        self.controls[device_id][control_data["href"]] = control_data
        self._call_ipc_method("control_register", device_id, control_data)

    def delControl(self, device_id, control_data):
        if device_id in self.controls:
            self.controls[device_id].pop(control_data["href"], None)
        self._call_ipc_method("control_unregister", device_id, control_data)

    def get_node_self(self, api_version="v1.1"):
        return self._call_ipc_method("self_get", api_version)

    def get_reg_status(self):
        return self._call_ipc_method("status_get")

    def addClock(self, clk_data):
        self._call_ipc_method("clock_register", clk_data)

    def updateClock(self, clk_data):
        self._call_ipc_method("clock_update", clk_data)

    def delClock(self, clk_name):
        self._call_ipc_method("clock_unregister", clk_name)

    def debug_message(self, code):
        msg = {FAC_SUCCESS: "Success!",
               FAC_EXISTS: "Service already exists",
               FAC_UNREGISTERED: "Service isn't yet registered",
               FAC_UNAUTHORISED: "Unauthorised",
               FAC_UNSUPPORTED: "Unsupported",
               FAC_OTHERERROR: "Other error"}[code]
        return msg



import websocket
import signal
import sys
import threading
import json

# This is a very thin wrapper around python WebSocketApp
# to allow easy use with threading by inheriting threading.Thread


class WebSocketClient(threading.Thread):
    daemon = True

    def __init__(self, wsAddr, sslopt=None):
        self.started = threading.Event()
        self.wsAddr = wsAddr
        self._keep_running = False
        threading.Thread.__init__(self)
        self.sslopt = sslopt

    def run(self):
        self._keep_running = True
        self.ws = websocket.WebSocketApp(self.wsAddr,
                                         on_message=self._on_message,
                                         on_error=self._on_error,
                                         on_close=self._on_close,
                                         on_open=self._on_open)
        while self._keep_running:
            self.__setstarted()
            self.ws.run_forever(sslopt=self.sslopt)

    def __setstarted(self):
        self.started.set()

    # These are just here to make the function signatures work
    # the user shouldn't be fiddling with the ws

    def _on_message(self, ws, message):
        self.onMessage(message)

    def _on_error(self, ws, error):
        self.onError(error)

    def _on_close(self, ws):
        self.onClose()

    def _on_open(self, ws):
        # Grab the websocket so we can use it to send later
        self.ws = ws
        self.onOpen()

    def onMessage(self, message):
        # over-ride this method in child class
        # to alter message handling behaviour
        pass

    def onError(self, error):
        # over-ride this method in child class
        # to alter error handling behaviour
        raise Exception(error)

    def onClose(self):
        # over-ride this method in child class
        # to alter actions when the websocket
        # is closed
        pass

    def onOpen(self):
        # over-ride this method in child class
        # to alter startup behaviour
        pass

    def sendJSON(self, message):
        self.ws.send(json.dumps(message))

    def sendPlain(self, message):
        self.ws.send(message)

    def stop(self):
        self._keep_running = False
        self.ws.close()


if __name__ == "__main__":  # pragma: no cover
    websocketClient = WebSocketClient("ws://localhost:8090/ws/")

    def signal_handler(rxsignal, frame):
        websocketClient.stop()
        sys.exit(0)
    signal.signal(signal.SIGINT, signal_handler)
    websocketClient.run()



#!/usr/bin/python
#
# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from setuptools import setup
import setuptools
import os
import sys
import json
from setuptools.command.develop import develop
from setuptools.command.install import install


def create_default_conf():
    fname = "/etc/nmoscommon/config.json"
    if os.path.isfile(fname):
        return

    # Defaults defined in nmoscommonconfig.py
    defaultData = {
        "priority": 0,
        "https_mode": "disabled",
        "prefer_ipv6": False,
        "prefer_hostnames": False,
        "node_hostname": None,
        "fix_proxy": "disabled",
        "logging": {
            "level": "DEBUG",
            "fileLocation": "/var/log/nmos.log",
            "output": [
                "file",
                "stdout"
            ]
        },
        "nodefacade": {
            "NODE_REGVERSION": "v1.2"
        },
        "oauth_mode": False,
        "dns_discover": True,
        "mdns_discover": True,
        "interfaces": []
    }

    try:
        try:
            os.makedirs(os.path.dirname(fname))
        except OSError as e:
            if e.errno != os.errno.EEXIST:
                raise
            pass

        with open(fname, 'w') as outfile:
            json.dump(defaultData,
                      outfile,
                      sort_keys=True,
                      indent=4,
                      separators=(",", ": "))
    except OSError as e:
        if e.errno != os.errno.EACCES:
            raise
        pass
        # Default config couldn't be created.
        # Code will fallback to hardcoded defaults.
        # This should only happen with un-priviladged installs


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        create_default_conf()


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        create_default_conf()


def is_package(path):
    return (
        os.path.isdir(path) and
        os.path.isfile(os.path.join(path, '__init__.py'))
        )


def find_packages(path, base=""):
    """ Find all packages in path """
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if is_package(dir):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages


packages = find_packages(".")
package_names = packages.keys()

# REMEMBER: If this list is updated, please also update stdeb.cfg
packages_required = [
    "gevent>=1.2.2",
    "greenlet>=0.4.13",
    "gevent-websocket>=0.9.3",
    "six>=1.10.0",
    "flask>=0.10.1",
    "jinja2>=2.7.2",
    "werkzeug>=0.14.1",
    "itsdangerous>=0.24",
    "socketio-client>=0.5.3",
    "flask-sockets>=0.1",
    "pyzmq>=15.2.0",
    "pygments>=1.6",
    "python-dateutil>=2.4.2",
    "oauthlib<3.0.0",
    "requests-oauthlib<1.2.0",
    "flask-oauthlib>=0.9.1",
    "ws4py>=0.3.2",
    "requests>=2.6",
    "jsonschema>=2.3.0",
    "netifaces>=0.10.6",
    "websocket-client>=0.18.0",
    "ujson>=1.33",
    "mediatimestamp>=1.0.1",
    "mediajson>=1.0.0",
    "authlib>=0.10",
    "pyopenssl",
    "dnspython>=1.12.0",
    "zeroconf-monkey>=0.1.0"
]

# Check if setuptools is recent enough to recognise python_version syntax
# Pybonjour is only available for Python 2
# wsaccel doesn't work with python 3.6
if int(setuptools.__version__.split(".", 1)[0]) < 18:
    # Check python version using legacy mechanism
    if sys.version_info[0:2] < (3, 0):
        packages_required.extend([
            "zeroconf<=0.19.1"
        ])  # Version locked to preserve Python 2 compatibility
    else:
        packages_required.extend([
            "zeroconf>=0.21.0"
        ])

    if sys.version_info[0:2] < (3, 6):
        packages_required.extend([
            'wsaccel>=0.6.2',
        ])
else:
    packages_required.extend([
        "zeroconf==0.19.1;python_version<'3.0'",
        "zeroconf>=0.21.0;python_version>='3.0'",
        'wsaccel>=0.6.2;python_version<"3.6"'
    ])

deps_required = []


setup(
    name="nmoscommon",
    version="0.19.15",
    description="Common components for the BBC's NMOS implementations",
    url='https://github.com/bbc/nmos-common',
    author='Peter Brightwell',
    author_email='peter.brightwell@bbc.co.uk',
    license='Apache 2',
    packages=package_names,
    package_dir=packages,
    install_requires=packages_required,
    dependency_links=deps_required,
    scripts=[],
    data_files=[],
    long_description="Common components for the BBC's NMOS implementations",
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    }
)



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zeroconf_monkey import ServiceInfo
from socket import inet_aton
import time
import six


class MDNSRegistration(object):

    def __init__(self, interfaces, name, regtype, port, txtRecord=None):
        if not txtRecord:
            txtRecord = {}
        self.name = name.replace(".", "-")
        self.regtype = regtype
        self.port = port
        self.txtRecord = self._conformTxtRecord(txtRecord)
        self.interfaces = interfaces
        self.info = {}

    def register(self):
        nameList = self._makeNamesUnique()
        regtype = self.regtype + ".local."
        for interface in self.interfaces:
            name = nameList[interface.ip] + "." + self.regtype + ".local."
            self.info[interface.ip] = ServiceInfo(
                regtype,
                name,
                port=self.port,
                properties=self.txtRecord,
                address=inet_aton(interface.ip)
            )
            interface.registerService(self.info[interface.ip])

    """TXT record entries must only be strings"""
    def _conformTxtRecord(self, record):
        conformedRecord = {}
        for key, value in six.iteritems(record):
            conformedRecord[key] = str(value)
        return conformedRecord

    def update(self, name=None, regtype=None, port=None, txtRecord=None):
        if name is not None:
            self.name = name
        if regtype is not None:
            self.regtype = regtype
        if port is not None:
            self.port = port
        if txtRecord is not None:
            self.txtRecord = self._conformTxtRecord(txtRecord)
        self.unRegister()
        time.sleep(1)  # Seems to fail without this sometimes :(
        self.register()

    def _makeNamesUnique(self):
        if len(self.interfaces) == 1:
            return {self.interfaces[0].ip: self.name}
        else:
            nameList = {}
            for interface in self.interfaces:
                # Zeroconf cannot deal with "." in names, so have to replace with dashes
                noDotIp = interface.ip.replace(".", "-")
                nameList[interface.ip] = self.name + "_{}".format(noDotIp)
            return nameList

    def unRegister(self):
        for interface in self.interfaces:
            interface.unregisterService(self.info[interface.ip])



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from geventwebsocket.handler import WebSocketHandler
from gevent.pywsgi import WSGIServer
import threading
import socket


class HttpServer(threading.Thread):

    daemon = True    # causes server to stop on main thread quitting

    def __init__(self, api, port=0, host='0.0.0.0', api_args=None, api_kwargs=None, ssl=None):
        self.api_args = api_args
        self.api_kwargs = api_kwargs
        if self.api_args is None:
            self.api_args = []
        if self.api_kwargs is None:
            self.api_kwargs = {}
        self.api_class = api
        self.api = None
        self.port = port
        self.host = host
        self.started = threading.Event()
        self.failed = None
        self.ssl = ssl

        threading.Thread.__init__(self)

    def run(self):
        try:
            self.api = self.api_class(*self.api_args, **self.api_kwargs)
            port = self.port
            while True:
                if self.port == 0:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    sock.bind(('localhost', 0))
                    self.port = sock.getsockname()[1]
                    sock.close()

                if self.port == 0:
                    print("ERROR! Port still 0!")
                    raise Exception("Port still 0 after socket binding")

                try:
                    # self.api.app.before_first_request = self.__setstarted()
                    if self.ssl is not None:
                        self.server = WSGIServer((self.host, self.port), self.api.app, handler_class=WebSocketHandler, **self.ssl)
                    else:
                        self.server = WSGIServer((self.host, self.port), self.api.app, handler_class=WebSocketHandler)

                    self.server.start()
                    self.__setstarted()
                    self.server.serve_forever()
                except socket.error as e:
                    if e.errno == 48:
                        if port != 0:
                            raise
                        else:
                            self.port = port
                            continue
                    else:
                        self.failed = e
                        self.__setstarted()
                break
        except Exception as e:
            self.failed = e
            if not self.started.is_set():
                self.__setstarted()

    def stop(self):
        self.api.stop()

    def __setstarted(self):
        if self.api is not None:
            try:
                self.api.port = self.port
            except Exception as e:
                self.failed = e
        self.started.set()
        if self.failed is None:
            print("HttpServer running")
        else:
            print("HttpServer failed to start")



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# DEPRECATED - This class is deprecated - please use timestamp.py

from __future__ import print_function
from __future__ import absolute_import

import ctypes
from .timestamp import Timestamp

try:
    import ipppython.ptptime
    IPP_PYTHON = True
except ImportError:
    IPP_PYTHON = False

__all__ = ["ptp_time", "ptp_detail"]


class timespec(ctypes.Structure):
    _fields_ = [
        ('tv_sec', ctypes.c_long),
        ('tv_nsec', ctypes.c_long)
    ]


# This method is deprecated, use Timestamp.get_time()
def ptp_data():
    t = timespec()
    ts = Timestamp.get_time()
    t.tv_sec = int(ts.to_nanosec() * 1e-9)
    t.tv_nsec = int(ts.to_nanosec() - (t.tv_sec * 1e9))
    return t


# This method is deprecated, use Timestamp.get_time()
if IPP_PYTHON:
    ptp_data = ipppython.ptptime.ptp_data  # noqa F811
    FD_TO_CLOCKID = ipppython.ptptime.FD_TO_CLOCKID


def ptp_time():
    t = ptp_data()
    return t.tv_sec + t.tv_nsec * 1e-9


def ptp_detail():
    t = ptp_data()
    return [t.tv_sec, t.tv_nsec]


if __name__ == "__main__":  # pragma: no cover
    print(ptp_time())
    print(ptp_detail())



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from authlib.jose import JWTClaims
from authlib.jose.errors import InvalidClaimError, MissingClaimError


class JWTClaimsValidator(JWTClaims):

    def __init__(self, payload, header, options=None, params=None):
        super(JWTClaimsValidator, self).__init__(payload, header, options, params)

    def validate_iss(self):  # placeholder
        super(JWTClaimsValidator, self).validate_iss()

    def validate_sub(self):  # placeholder
        super(JWTClaimsValidator, self).validate_sub()

    def validate_aud(self):  # placeholder
        super(JWTClaimsValidator, self).validate_aud()

    def validate_nmos(self):
        claim_name = "x-nmos-api"
        valid_claim_option = self.options.get(claim_name)
        valid_claim_value = valid_claim_option.get('value')

        actual_claim_value = self.get(claim_name)
        if not actual_claim_value:
            raise MissingClaimError(claim_name)
        if not valid_claim_value:
            return

        # TODO - add validation for versions and read/write
        for attr in valid_claim_value:
            if actual_claim_value.get(attr) != valid_claim_value.get(attr):
                raise InvalidClaimError(claim_name)

    def validate(self, now=None, leeway=0):
        super(JWTClaimsValidator, self).validate()
        self.validate_nmos()



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from six.moves import socketserver
from code import InteractiveConsole

interpreter_globals = {}


class InteractiveServer(socketserver.BaseRequestHandler):
    def handle(self):
        global interpreter_globals

        file = self.request.makefile()
        shell = Shell(file, locals=interpreter_globals)
        try:
            shell.interact()
        except SystemExit:
            pass


class Shell(InteractiveConsole):
    def __init__(self, file, *args, **kwargs):
        self.file = sys.stdout = file
        InteractiveConsole.__init__(self, *args, **kwargs)
        return

    def write(self, data):
        self.file.write(data)
        self.file.flush()

    def raw_input(self, prompt=""):
        self.write(prompt)
        return self.file.readline()


def interact(address=("0.0.0.0", 9999)):
    server = socketserver.TCPServer(address, InteractiveServer)
    server.serve_forever()


port = 9999
if __name__ == '__main__':  # pragma: no cover
    if len(sys.argv) > 1:
        port = int(sys.argv[1])
    interact(("0.0.0.0", port))
